 #version 300 es

precision highp float;

#define PI 3.14159265

// 定义 散射系数和闪射密度，用于计算大气层中的瑞利散射。
uniform vec3 ScatterRayleigh;
uniform float  HDensityRayleigh;

// 定义散射系数、散射不对称性、吸收系数和散射密度，用于计算大气层中的米氏散射
uniform float ScatterMie;
uniform float AsymmetryMie;
uniform float AbsorbMie;
uniform float HDensityMie;

// 定义臭氧吸收系数、臭氧层中心高度和臭氧层厚度，用于计算大气层中的臭氧吸收。
uniform vec3 AbsorbOzone;
uniform float  OzoneCenterHeight;
uniform float OzoneThickness;

// 定义行星半径和大气层半径。
uniform float PlanetRadius;
uniform float AtmosphereRadius;

uniform sampler2D Transmittance;

uniform vec3 AtmosEyePosition;
uniform int MarchStepCount;
uniform vec3 SunDirection;
uniform vec3 SunIntensity;
// uniform vec2 uResolution;

uniform float POSTCOLOR_A;
uniform float POSTCOLOR_B;
uniform float POSTCOLOR_C;
uniform float POSTCOLOR_D;
uniform float POSTCOLOR_E;

// in vec2 texCoord;
in vec2 vUv;
out vec4 o_FragColor;

// 色调映射函数，用于调整颜色的亮度和对比度
vec3 tonemap(vec3 param)
{
    return (param * (vec3(POSTCOLOR_A) * param + vec3(POSTCOLOR_B)))
    / (param * (vec3(POSTCOLOR_C) * param + vec3(POSTCOLOR_D)) + vec3(POSTCOLOR_E));
}

// 后处理函数，用于添加随机噪点和调整颜色的亮度和对比度。
vec3 postProcessColor(vec2 seed, vec3 param)
{
    param = tonemap(param);

    float rand = fract(sin(dot(seed, vec2(12.9898, 78.233) * vec2(2.0))) * 43758.5453);
    param = vec3(255.) * clamp(pow(param, vec3(1. / 2.2)), 0., 1.);
    param.x = rand < (param.x - floor(param.x)) ? ceil(param.x) : floor(param.x);
    param.y = rand < (param.y - floor(param.y)) ? ceil(param.y) : floor(param.y);
    param.z = rand < (param.z - floor(param.z)) ? ceil(param.z) : floor(param.z);

    return param / vec3(255.);
}

// 判断光线与地球是否有交点
bool hasIntersectionWithSphere(vec3 o, vec3 d, float R)
{
    float A = dot(d, d);
    float B = 2.*dot(o, d);
    float C = dot(o, o) - R*R;
    float delta = B*B - 4.*A*C;
    return (delta>=0.)&&((C<=0.)||(B<=0.));
}

// 找到一条射线与一个圆相交的最近点
bool findClosestIntersectionWithCircle(vec2 o, vec2 d, float R, out float t){
    float A = dot(d, d);
    float B = 2. * dot(o, d);
    float C = dot(o, o) - R*R;
    float delta = B*B - 4.*A*C;
    if (delta < 0.){
        return false;
    }
    t = (-B + (C<=0.? sqrt(delta):-sqrt(delta)))/(2.*A);
    return (C<=0.)||(B<=0.);
}

// 计算大气层中的散射和吸收系数
void getSigmaST(float h, out vec3 sigmaS, out vec3 sigmaT)
{
    vec3 rayleigh = ScatterRayleigh * vec3(exp(-h / HDensityRayleigh));

    float mieDensity = exp(-h / HDensityMie);
    float mieS = ScatterMie * mieDensity;
    float mieT = (ScatterMie + AbsorbMie) * mieDensity;

    vec3 ozone = AbsorbOzone *
    vec3(max(0.0f, 1. - 0.5 * abs(h - OzoneCenterHeight) / OzoneThickness));

    sigmaS = rayleigh + vec3(mieS);
    sigmaT = rayleigh + vec3(mieT) + ozone;
}

// 计算散射相函数  h表示当前点到地球表面的距离，u表示光线的方向向量与垂直于地球表面的向量之间的夹角的余弦值。
vec3 evalPhaseFunction(float h, float u)
{
    vec3 sRayleigh = ScatterRayleigh * vec3(exp(-h / HDensityRayleigh));
    float sMie = ScatterMie * exp(-h / HDensityMie);
    // 将Rayleigh和Mie分量相加，得到总的散射光线强度
    vec3 s = sRayleigh + vec3(sMie);

    float g = AsymmetryMie, g2 = g * g, u2 = u * u;
    float pRayleigh = 3. / (16. * PI) * (1. + u2);

    float m = 1. + g2 - 2. * g * u;
    float pMie = 3./ (8. * PI) * (1. - g2) * (1. + u2) / ((2. + g2) * m * sqrt(m));

    vec3 result;
    result.x = s.x > 0. ? (pRayleigh * sRayleigh.x + pMie * sMie) / s.x : 0.;
    result.y = s.y > 0. ? (pRayleigh * sRayleigh.y + pMie * sMie) / s.y : 0.;
    result.z = s.z > 0. ? (pRayleigh * sRayleigh.z + pMie * sMie) / s.z : 0.;
    // x、y、z分别表示散射光线在RGB三个通道上的强度。
    return result;
}

// 计算透射率
vec3 getTransmittance(sampler2D T, float h, float theta)
{
    float u = h / (AtmosphereRadius - PlanetRadius);
    float v = 0.5 + 0.5 * sin(theta);
    return texture(T, vec2(u, v),0.).rgb;
}

// 计算每个步长的散射和吸收，以及透射率。
// phaseU表示散射相函数的参数，
// ori表示光线的起点，
// dir表示光线的方向向量，
// thisT和nextT表示当前步长和下一个步长的距离，
// sumSigmaT和inScattering分别表示累计的吸收系数和散射光线的强度。
void marchStep(float phaseU, vec3 ori, vec3 dir, float thisT, float nextT, inout vec3 sumSigmaT, inout vec3 inScattering, vec3 sunDir)
{
    // 当前步长和下一个步长的中间点的位置posR
    float midT = 0.5 * (thisT + nextT);
    vec3 posR = vec3(0, ori.y + PlanetRadius, 0.) + dir * vec3(midT);
    float h    = length(posR) - PlanetRadius;

    // 计算当前点的散射系数和吸收系数
    vec3 sigmaS, sigmaT;
    getSigmaST(h, sigmaS, sigmaT);

    // 当前步长的吸收系数deltaSumSigmaT，即当前步长内光线被吸收的强度。
    vec3 deltaSumSigmaT = vec3(nextT - thisT) * sigmaT;
    // 当前步长内的透射率，即光线在当前步长内没有被吸收的概率。这个值是通过累计的吸收系数和当前步长的吸收系数计算得到的。
    vec3 eyeTrans = exp(-sumSigmaT - vec3(0.5) * deltaSumSigmaT);

    // 计算了光线与太阳方向的夹角sunTheta
    float sunTheta = PI / 2. - acos(dot(-sunDir, normalize(posR)));

    // 判断当前步长内的光线是否与地球相交。如果没有相交，则计算当前步长内的散射光线的强度。
    if (!hasIntersectionWithSphere(posR, -sunDir, PlanetRadius))
    {
        vec3 rho = evalPhaseFunction(h, phaseU);
        vec3 sunTrans = getTransmittance(Transmittance, h, sunTheta);
        inScattering += vec3(nextT - thisT) * eyeTrans * sigmaS * rho * sunTrans;
    }

    sumSigmaT += deltaSumSigmaT;
}

vec3 calSunDir(vec2 sunAngle){
    vec2 sunRad = (sunAngle) * vec2(PI / 180., -PI / 180.);
    float x = cos(sunRad.x) * cos(sunRad.y);
    float y = sin(sunRad.y);
    float z = sin(sunRad.x) * cos(sunRad.y);
    return normalize(vec3(x, y, z));
}

//     texCoord 0:[0,0]; 1:[2,0]; 2:[0,2]; 3:[2,2];
void main()
{

    vec2 texCoord = vUv;
    texCoord.x = 0.5;

    // 当前像素在水平方向的角度   texCoord.x：当前像素在水平方向上的纹理坐标
    float phi = 2. * PI * texCoord.x;
    // 当前像素在垂直方向上的位置    texCoord.y：当前像素在垂直方向上的纹理坐标
    float vm = texCoord.y * 1.1015625 - 0.1015625;
    // 当前像素在垂直方向上的角度  `sign`是一个函数，返回参数的符号，即正数返回1，负数返回-1，0返回0
    float theta = sign(vm) * (PI / 2.) * vm * vm;
    float sinTheta = sin(theta), cosTheta = cos(theta);

    // 相机的位置
    vec3 ori = AtmosEyePosition;
    // 从相机位置出发到当前像素的方向   具体的计算方式是将球面坐标系转换为笛卡尔坐标系。
    vec3 dir = vec3(cos(phi) * cosTheta, sinTheta, sin(phi) * cosTheta);

    // 行星表面上的起点
    vec2 planetOri = vec2(0., ori.y + PlanetRadius);
    // 从起点出发到当前像素的方向
    vec2 planetDir = vec2(cosTheta, sinTheta);

    // find end point
    // 表示从起点出发到当前像素的射线与行星或大气层的交点距离。
    float endT = 0.;
    if (!findClosestIntersectionWithCircle(planetOri, planetDir, PlanetRadius, endT))
    {
        findClosestIntersectionWithCircle(planetOri, planetDir, AtmosphereRadius, endT);
    }

    // SunDirection
    vec3 sunDir = calSunDir(vec2(
        SunDirection.x,
        mix(0., 180., vUv.x)
    ));

    // phase function input
    // `-dir`表示当前像素到相机位置的方向，因为散射相函数的输入参数是入射方向和散射方向的夹角，所以这里需要将入射方向取反。
    float phaseU = dot(sunDir, -dir);

    // ray march
    // 当前射线的起点到当前像素的距离
    float distance = 0.;
    vec3 inScatter = vec3(0., 0., 0.);
    vec3 sumSigmaT = vec3(0., 0., 0.);

    float dt = (endT - distance) / float(MarchStepCount);
    for (int i = 0; i < MarchStepCount; ++i)
    {
        float nextT = distance + dt;
        marchStep(phaseU, ori, dir, distance, nextT, sumSigmaT, inScatter, sunDir);
        distance = nextT;
    }

    // 散射颜色 * 太阳强度
    vec3 skyColor = inScatter * SunIntensity;
    skyColor = postProcessColor(texCoord, skyColor);
    o_FragColor = vec4(skyColor, 1.);
    // o_FragColor = vec4(texCoord, 0., 1.);
    // o_FragColor = vec4(phi, vm, 0., 1.);

    // o_FragColor = vec4(sunDir, 1.);
}